/**********************************************************************
 * VPI Examples to read task/function argument values -- PLI
 * application using VPI routines
 *
 * C source for functions to the logic value of system task/function
 * arguments.
 *
 * NOTE: THESE ROUTINES A PROVIDED AS SHORT EXAMPLES ON HOW TO ACCESS
 * TASK/FUNCTION ARGUMENTS, AND HOW TO PERFORM ERROR CHECKING ON IF
 * THE DESIRED INFORMATION WAS OBTAINED.  THESE EXAMPLES ARE NOT
 * INTENDED TO BE COMPLETE PLI APPLICATION IN AND OF THEMSELVES.
 *
 * Usage:
 *   arg_val = PLIbook_get_arg_int_val_vpi(arg_index_number);
 *     Returns a value as an int for a system task/function argument,
 *     using the index number of the argument, beginning with 1.
 *
 *   arg_val = PLIbook_get_arg_real_val_vpi(arg_index_number);
 *     Returns a value as a double for a system task/function arg,
 *     using the index number of the argument, beginning with 1.
 *
 *   arg_val = PLIbook_get_arg_string_val_vpi(arg_index_number);
 *     Returns a value as a string pointer for a system task/func arg,
 *     using the index number of the argument, beginning with 1.
 *
 * For the book, "The Verilog PLI Handbook" by Stuart Sutherland
 *  Copyright 1999 & 2001, Kluwer Academic Publishers, Norwell, MA, USA
 *   Contact: www.wkap.il
 *  Example copyright 1998, Sutherland HDL Inc, Portland, Oregon, USA
 *   Contact: www.sutherland-hdl.com
 *********************************************************************/

#define PLIbookDebug 1 /* set to non-zero for verbose debug messages */

#include <stdlib.h>    /* ANSI C standard library */
#include <stdio.h>     /* ANSI C standard input/output library */
#include <stdarg.h>    /* ANSI C standard arguments library */
#include "vpi_user.h"  /* IEEE 1364 PLI VPI routine library  */

#include "get_arg_handle_vpi_efficient.c"  /* example from Chapter 4 */

/* Prototypes of the utility applications */
int     PLIbook_get_arg_int_val_vpi(int argNum);
double  PLIbook_get_arg_real_val_vpi(int argNum);
char   *PLIbook_get_arg_string_val_vpi(int argNum);


/**********************************************************************
 * PLIbook_get_arg_int_val_vpi()
 * Return the value of a system task/function argument as an integer.
 * Similar to acc_fetch_tfarg_int()
 *********************************************************************/
int PLIbook_get_arg_int_val_vpi(int argNum)
{
  vpiHandle    arg_h;
  s_vpi_value  argVal;
  s_vpi_error_info err;  /* structure for error handling */

  arg_h = PLIbook_get_arg_handle_vpi(argNum);
  if (arg_h == NULL) {
    vpi_printf("ERROR: PLIbook_get_arg_int_val_vpi() could not obtain arg handle\n");
    return(0);
  }
  argVal.format = vpiIntVal;
  vpi_get_value(arg_h, &argVal);
  if (vpi_chk_error(&err)) {
    #if PLIbookDebug  /* if error, generate verbose debug message */
      vpi_printf("ERROR: PLIbook_get_arg_int_val_vpi() could not obtain arg value\n");
      vpi_printf("File %s, Line %d: %s\n", err.file, err.line,
                  err.message);
    #endif
    return(0);
  }
  return(argVal.value.integer);
}


/**********************************************************************
 * PLIbook_get_arg_real_val_vpi()
 * Return the value of a system task/function argument as a double.
 * Similar to acc_fetch_tfarg()
 *********************************************************************/
double PLIbook_get_arg_real_val_vpi(int argNum)
{
  vpiHandle arg_h;
  s_vpi_value argVal;
  s_vpi_error_info err;  /* structure for error handling */

  arg_h = PLIbook_get_arg_handle_vpi(argNum);
  if (arg_h == NULL) {
    vpi_printf("ERROR: PLIbook_get_arg_real_val_vpi() could not obtain arg handle\n");
    return(0);
  }
  argVal.format = vpiRealVal;
  vpi_get_value(arg_h, &argVal);
  if (vpi_chk_error(&err)) {
    #if PLIbookDebug  /* if error, generate verbose debug message */
      vpi_printf("ERROR: PLIbook_get_arg_real_val_vpi() could not obtain arg value\n");
      vpi_printf("File %s, Line %d: %s\n", err.file, err.line,
                  err.message);
    #endif
    return(0.0);
  }
  return(argVal.value.real);
}


/**********************************************************************
 * PLIbook_get_arg_string_val_vpi()
 * Return the value of a system task/function argument as a string.
 * Similar to acc_fetch_tfarg_str()
 *********************************************************************/
char *PLIbook_get_arg_string_val_vpi(int argNum)
{
  vpiHandle arg_h;
  s_vpi_value argVal;
  s_vpi_error_info err;  /* structure for error handling */
  static char null_string[1] = {'\0'};

  arg_h = PLIbook_get_arg_handle_vpi(argNum);
  if (arg_h == NULL) {
    vpi_printf("ERROR: PLIbook_get_arg_string_val_vpi() could not obtain arg handle\n");
    return(0);
  }
  argVal.format = vpiStringVal;
  vpi_get_value(arg_h, &argVal);
  if (vpi_chk_error(&err)) {
    #if PLIbookDebug  /* if error, generate verbose debug message */
      vpi_printf("ERROR: PLIbook_get_arg_string_val_vpi() could not obtain arg value\n");
      vpi_printf("File %s, Line %d: %s\n", err.file, err.line,
                  err.message);
    #endif
    return(null_string);
  }
  return(argVal.value.str);
}

/*********************************************************************/
