/**********************************************************************
 * $read_vecval example -- PLI application using VPI routines
 *
 * C source to read logic values of a net and print the net name and
 * current logic value.
 *
 * Usage: $read_vecval(<net_name>);
 *
 * For the book, "The Verilog PLI Handbook" by Stuart Sutherland
 *  Copyright 1999 & 2001, Kluwer Academic Publishers, Norwell, MA, USA
 *   Contact: www.wkap.il
 *  Example copyright 1998, Sutherland HDL Inc, Portland, Oregon, USA
 *   Contact: www.sutherland-hdl.com
 *********************************************************************/

#define VPI_1995 0 /* set to non-zero for Verilog-1995 compatibility */

#include <stdlib.h>    /* ANSI C standard library */
#include <stdio.h>     /* ANSI C standard input/output library */
#include <stdarg.h>    /* ANSI C standard arguments library */
#include "vpi_user.h"  /* IEEE 1364 PLI VPI routine library  */

#if VPI_1995
#include "../vpi_1995_compat.h"  /* kludge new Verilog-2001 routines */
#endif


/* prototypes of routines in this PLI application */
PLI_INT32  PLIbook_ReadVecVal_calltf(PLI_BYTE8 *user_data);
PLI_INT32  PLIbook_ReadVecVal_compiletf(PLI_BYTE8 *user_data);
int        PLIbook_getbit(PLI_INT32 word, int bit_num);
char       PLIbook_get_4state_val(int aval, int bval);

/**********************************************************************
 * VPI Registration Data
 *********************************************************************/
void PLIbook_ReadVecVal_register()
{
  s_vpi_systf_data tf_data;
  tf_data.type        = vpiSysTask;
  tf_data.sysfunctype = 0;
  tf_data.tfname      = "$read_vecval";
  tf_data.calltf      = PLIbook_ReadVecVal_calltf;
  tf_data.compiletf   = PLIbook_ReadVecVal_compiletf;
  tf_data.sizetf      = NULL;
  tf_data.user_data   = NULL;
  vpi_register_systf(&tf_data);
}
/*********************************************************************/

/**********************************************************************
 * compiletf application
 *********************************************************************/
PLI_INT32 PLIbook_ReadVecVal_compiletf(PLI_BYTE8 *user_data)
{
  vpiHandle systf_h, arg_itr, arg_h;
  PLI_INT32  arg_type;

  /* obtain a handle to the system task instance */
  systf_h = vpi_handle(vpiSysTfCall, NULL);
  if (systf_h == NULL) {
    vpi_printf("ERROR: $read_vecval failed to obtain systf handle\n");
    vpi_control(vpiFinish, 1);  /* abort simulation */
    return(0);
  }

  /* obtain handles to system task arguments */
  arg_itr = vpi_iterate(vpiArgument, systf_h);
  if (arg_itr == NULL) {
    vpi_printf("ERROR: $read_vecval requires 1 argument\n");
    vpi_control(vpiFinish, 1);  /* abort simulation */
    return(0);
  }

  /* check the type of object in system task arguments */
  arg_h = vpi_scan(arg_itr);
  arg_type = vpi_get(vpiType, arg_h);
  if (arg_type != vpiNet && arg_type != vpiReg) {
    vpi_printf("ERROR: $read_vecval arg must be a net or reg\n");
    vpi_free_object(arg_itr); /* free iterator memory */
    vpi_control(vpiFinish, 1);  /* abort simulation */
    return(0);
  }

  /* check that there are no more system task arguments */
  arg_h = vpi_scan(arg_itr);
  if (arg_h != NULL) {
    vpi_printf("ERROR: $read_vecval can only have 1 argument\n");
    vpi_free_object(arg_itr); /* free iterator memory */
    vpi_control(vpiFinish, 1);  /* abort simulation */
    return(0);
  }
  return(0);  /* no syntax errors detected */
}

/**********************************************************************
 * calltf routine
 *********************************************************************/
PLI_INT32 PLIbook_ReadVecVal_calltf(PLI_BYTE8 *user_data)
{
  vpiHandle   systf_h, arg_itr, arg_h, vector_h;
  s_vpi_value vector_val;     /* structure to receive vector value */
  PLI_INT32   vector_size, array_size;
  int  i, bit_num, avalbit, bvalbit;
  char vlogval;

  /* obtain a handle to the system task instance */
  systf_h = vpi_handle(vpiSysTfCall, NULL);

  /* obtain handle to system task argument
     compiletf has already verified only 1 arg with correct type */
  arg_itr = vpi_iterate(vpiArgument, systf_h);
  vector_h = vpi_scan(arg_itr);
  vpi_free_object(arg_itr);  /* free iterator memory */

  vector_size = vpi_get(vpiSize, vector_h); /* determine number of...*/
  array_size  = ((vector_size-1) / 32 + 1); /* ...elements in array  */

  vector_val.format = vpiVectorVal;        /* set value format field */

  vpi_get_value(vector_h, &vector_val); /* read vector's logic value */

  vpi_printf("\nVector %s encoded value:\n",
             vpi_get_str(vpiName,vector_h));
  for (i=0; i<array_size; i++) {
    /* for simplicity, the following loop assumes the Verilog LSB    */
    /* is bit 0.  More robust code would obtain handles for the      */
    /* vpiRightRange and vpiLeftRange expressions and determine the  */
    /* vector declaration                                            */
    for (bit_num=0; bit_num<=31; bit_num++) {
      avalbit=PLIbook_getbit(vector_val.value.vector[i].aval, bit_num);
      bvalbit=PLIbook_getbit(vector_val.value.vector[i].bval, bit_num);
      vlogval=PLIbook_get_4state_val(avalbit, bvalbit);
      vpi_printf("  bit[%2d]  aval/bval = %d/%d  4-state value = %c\n",
                 (i*32+bit_num), avalbit, bvalbit, vlogval);
      /* quit when reach last bit of Verilog vector */
      if ((i*32+bit_num) == vector_size-1) break;
    }
  }
  return(0);
}

/**********************************************************************
 * Function to determine if a specific bit is set in a 32-bit word.
 * Sets the least-significant bit of a mask value to 1 and shifts the
 * mask left to the desired bit number.
 *********************************************************************/
int PLIbook_getbit(PLI_INT32 word, int bit_num)
{
  int mask;
  mask = 0x00000001 << bit_num;
  return((word & mask)? 1: 0);  /* 1 == TRUE, 0 == FALSE */
}

/**********************************************************************
 * Function to convert aval/bval encoding to 4-state logic represented
 * as a C character.
 *********************************************************************/
char PLIbook_get_4state_val(int aval, int bval)
{
  if      (!bval && !aval) return('0');
  else if (!bval &&  aval) return('1');
  else if ( bval && !aval) return('z');
  else                     return('x');
}

/*********************************************************************/
