/**********************************************************************
 * VPI 1995 Compatibilty Kludges -- PLI application using VPI routines
 *
 * C source for functions to implement a few Verilog-2001 VPI routines
 * that did not exist in the Verilog-1995 standard.
 *
 * NOTE: THESE ROUTINES ARE INTENDED AS A TEMPORARY WORK AROUND FOR
 * TESTING PURPOSES ONLY.  THEY ARE KLUDGES FOR SIMULATORS THAT HAVE
 * NOT YET IMPLEMENTED THE FULL IEEE 1364-2001 STANDARD.  THE ROUTINES
 * USE THE SAME NAME AS THE IEEE STANDARD, AND THEREFORE WILL CONFLICT
 * WITH SIMULATORS THAT HAVE IMPLEMENTED THE STARDARD ROUTINES.
 *
 * Usage:
 *   vpi_control(operation_constant, ...);
 *     Controls certain simulation functions.  The only operation
 *     constants implemented are vpiStop and vpiFinish.  Other
 *     constants are ignored.
 *
 *   vpi_put_userdata(systf_handle, value);
 *     Stores a value in a work area that is unique to each instance
 *     of a system task/function.
 *
 *   value = vpi_get_userdata(systf_handle);
 *     Retreives the value stored in the work area for a specific
 *     instance of a system task/function.
 *
 * For the book, "The Verilog PLI Handbook" by Stuart Sutherland
 *  Copyright 1999 & 2001, Kluwer Academic Publishers, Norwell, MA, USA
 *   Contact: www.wkap.il
 *  Example copyright 2001, Sutherland HDL Inc, Portland, Oregon, USA
 *   Contact: www.sutherland-hdl.com
 *********************************************************************/

#include <stdlib.h>    /* ANSI C standard library */
#include <stdio.h>     /* ANSI C standard input/output library */
#include <stdarg.h>    /* ANSI C standard arguments library */
#include "vpi_user.h"  /* IEEE 1364 PLI VPI routine library  */
#include "veriuser.h"  /* IEEE 1364 PLI TF routine library */

#define PLIbookDebug 1 /* set to non-zero for verbose debug messages */

#define CONTROL  1  /* set to non-zero for vpi_control() kludge */
#define USERDATA 0  /* set to non-zero for vpi_put/get_userdata() kludge */

#if CONTROL
PLI_INT32  vpi_control(PLI_INT32 operation, ...);

/**********************************************************************
 * vpi_control()
 *   Allows a PLI application to control certain aspects of simulation.
 *   - vpi_control(vpiStop, n) is mapped to tf_dostop();
 *   - vpi_control(vpiFinish, n) is mapped to tf_dofinish();
 *
 *   This routine does not implement the all the features of the
 *   vpi_control() defined in the Verilog-2001 standard.  It only
 *   implements two of the control constants, does not implement the
 *   arguments to the controls, and always returns a 1, without
 *   checking to see if the functionality was successful.
 *********************************************************************/

PLI_INT32 vpi_control(PLI_INT32 operation, ...) {
  #define vpiStop                66 /* execute simulator's $stop */
  #define vpiFinish              67 /* execute simulator's $finish */
  #define vpiReset               68 /* execute simulator's $reset */
  #define vpiSetInteractiveScope 69 /* set simulator's interactive scope */

  switch(operation) {
    case vpiStop:     tf_dostop();   /* halt simulation */
                      break;
    case vpiFinish:   tf_dofinish(); /* abort simulation */
                      break;
  }
  return(1);
}
#endif  /* conditional compilation for vpi_control() */


#if USERDATA

void      *vpi_get_userdata(vpiHandle obj);
PLI_INT32  vpi_put_userdata (vpiHandle systf_h, void *userdata);

/**********************************************************************
 * vpi_put_userdata()
 *   Stores a data value in the work area for a specific system
 *   task/function instance.  Emulates the vpi_put_userdata() added in
 *   the Verilog-2001 standard.  Similar to tf_setworkarea().
 *
 *   This routine automatically allocates a work area that is specific
 *   to an instance of a system task or system function. The work area
 *   is created using a simple LIFO stack so that a work area can be
 *   allocated for any number of system task/function instances. Only
 *   one work area per task/function instance will be allocated.
 *
 *   The work area stores a void* pointer in its data field. The data
 *   can be retrieved using vpi_get_userdata().
 *
 *   NOTE: no thought or effort was put into memory efficiency or
 *   simulation performance.  This function is only provided as a
 *   kludge for simulators that have not yet implented the full
 *   Verilog-2001 standard.
 *********************************************************************/

/* Work area structure definition */
typedef struct PLIbook_userdata *PLIbook_userdata_p;
typedef struct PLIbook_userdata {
  vpiHandle  systf_h; /* shows which systf instance owns this space */
  void      *data;    /* data to be stored in userdata */
  PLIbook_userdata_p next_PLIbook_userdata;
} PLIbook_userdata_s;

/* allocate a global stack pointer for all task/function instances */
static PLIbook_userdata_p PLIbook_userdata_stack = NULL;


PLI_INT32 vpi_put_userdata (vpiHandle systf_h, void *userdata)
{
  PLIbook_userdata_p  tfinst_userdata;

  /* get top-of-stack pointer from global stack pointer */
  tfinst_userdata = PLIbook_userdata_stack;

  /* locate the userdata in the stack for this task instance */
  if (tfinst_userdata != NULL) {
    while (tfinst_userdata &&
           !vpi_compare_objects(tfinst_userdata->systf_h, systf_h) ) {
      tfinst_userdata = tfinst_userdata->next_PLIbook_userdata;
    }
  }

  /* if no work area found for this task instance, create one */
  if (tfinst_userdata == NULL) {
    tfinst_userdata = (PLIbook_userdata_p)malloc(sizeof(PLIbook_userdata_s));
    tfinst_userdata->systf_h = systf_h; /* set owner of this userdata */
    if (PLIbook_userdata_stack == NULL) {
      /* work area stack doesn't exist yet, create first location */
      tfinst_userdata->next_PLIbook_userdata = NULL;
      PLIbook_userdata_stack = tfinst_userdata;
    }
    else {
      tfinst_userdata->next_PLIbook_userdata = PLIbook_userdata_stack;
      PLIbook_userdata_stack = tfinst_userdata; /* set new top-of-stack */
    }
  }
  if (tfinst_userdata == NULL) {
    #ifdef PLIbookDebug  /* generate verbose debug message */
      vpi_printf("ERROR: failed to allocate userdata for this task instance\n");
    #endif
    return(0);  /* return 0 for error */
  }

  /* store user's data in the userdata area */
  tfinst_userdata->data = userdata;
  return(1);  /* return 1 for successful */
}

/**********************************************************************
 * vpi_get_userdata()
 *   Returns the data stored in the work area for a specific system
 *   task/function instance. Similar to tf_getworkarea().  A NULL is
 *   returned if unsuccessful.
 *********************************************************************/
void *vpi_get_userdata(vpiHandle systf_h)
{
  PLIbook_userdata_p  tfinst_userdata;

  /* locate the userdata in the stack for this task instance */
  tfinst_userdata = PLIbook_userdata_stack;  /* top-of-stack pointer */
  if (tfinst_userdata != NULL) {
    while (tfinst_userdata &&
           !vpi_compare_objects(tfinst_userdata->systf_h, systf_h) ) {
      tfinst_userdata = tfinst_userdata->next_PLIbook_userdata;
    }
  }
  if (tfinst_userdata == NULL) {
    #ifdef PLIbookDebug  /* generate verbose debug message */
      vpi_printf("Warning: no userdata storage found for this task instance\n");
    #endif
    return(NULL);
  }
  return(tfinst_userdata->data);
}
#endif  /* end of conditional compilation for vpi_put/get_userdata() */

/*********************************************************************/
